
//*****************************************************************************
//
// File Name	: 'SED1530.c'
// Title		: SPI interface driver for the E0855-2 SED1530 graphic LCD
// Author		: Marc Meise
// Created		: Oktober 2007
// Version		: Beta
// Target MCU	: Atmel AVR series
//
//
// This code is distributed under the GNU Public License
// which can be found at http://www.gnu.org/licenses/gpl.txt
//
//*****************************************************************************

/*  
	Disclaimer:
	This code is highly beta and might lack some funcionality and/or documentation.
	Furthermore the code has not been properly tested.
	Nevertheless, please expect most functions to work.
	01.12.2008 Fixed a bug which affected the "end of line"/newline determination.
			   Thanks to Klaus W. from mikrocontroller.net
			   -Added a simple itoa-function
	22.01.2008 Minor fixes in the hardware-SPI-prescaler code.
			   Tightend character set and shifted it to EEPROM (now fits into Mega8 and greater).
	21.01.2008 Added function "SED_7segment"
			   Shifted 7-segment definitions into EEPROM
	05.12.2007 Added support for software-SPI
	05.12.2007 Added function SED_pixel as a (very) basic graphic routine
	30.11.2007 Made CS and /CS redundant by switch "CS_ON"
	30.11.2007 Added "SED_contrast" function
	29.11.2007 Replaced the font file which was modified and debuged by Werner Boellmann


	I will not take any responsibility or what so ever for the functionality of the E0855-2 SED1530 driver.
	If you wreck anything - it's your own business...
	Modify this code and contribute freely.
*/

#include <avr/io.h>
#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <font.h>
#include <string.h>
#include <PICS.h>

// Port-definitions
// Alter the Port-definitions to your convenience.

#define LCDPORT		PORTB				//PORT where the GLCD ist connected
#define LCDDIR		DDRB				//LCD-Direction of LCD-PORT
#define SS			PB4					//Slave-Select (will only be set to output)
#define A0			PB0					//A0
#define RES 		PB3					//Reset
#define CS1 		PB1					//CS1 is Low-active
#define CS2 		PB2					//CS2 is high-active
#define MOSI		PB5					//MOSI
#define SCK			PB7					//SCK
#define CS_ON       1					//If CS and /CS are hardwired (not connected to the Controller, then set to '0')
static unsigned char	SPICLK=8;		//SPI-Prescaler (must be 2,8,16,32,64 or 128)

//Software-SPI

static unsigned char	SOFT_SPI=0;		// Use Software SPI (1=enable; 0=disable software-SPI)
#define SOFT_SPI_PORT 	PORTB			// Port for Software-SPI
#define SOFT_MOSI   	PB5				// Portpin MOSI for software-SPI
#define SOFT_SCK    	PB7				// Portpin SCK for software-SPI


// Only alter the following code, if you know what you're doing...

#define COL_Hi		0b00010001			//Column 16 HI is left for E0855-2	- Set the visible area		
#define COL_Lo		0b00000000			//Column 00 Lo is left for E0855-2
#define SC_width	96					//Screenwidth (96 for the E0855-2)

#define A0_Hi		PORTB|=(1<<A0)		//Change A0 SED1530 command-mode
#define A0_Lo	  	PORTB&=~(1<<A0)		//Change A0 SED1530 data-mode
#define RES_On  	PORTB&=~(1<<RES)	//Reset SED1530
#define RES_Off  	PORTB|=(1<<RES)		//Set SED1530


// Function prototypes

void Init_SPI (void);
void SPI_out (uint8_t Data);
void Init_SED1530 (void);
void SED_clear (void);
void SED_clear_page (uint8_t Page);
void SED_Chg_page (uint8_t Page);
void SED_Home_Col (void);
void Goto_XY (uint8_t X, uint8_t Y);
void SED_putc (char character);
void SED_puts (char *string);
void SED_puts_p (char *s_progmem);
void SED_7segment(uint8_t segno, uint8_t number);
void SED_symbol(uint8_t symbol, uint8_t OnOff);
void SED_newline(void);
void SED_showPIC (uint8_t picture);
void SED_puts_num (uint16_t numeric);
uint8_t ReverseBitorder (uint8_t Bits);
void SED_set_XY(uint8_t Page, uint8_t Column);
void SED_sleep (void);
void SED_wakeup (void);
void SED_contrast (uint8_t contrast);
void SED_set_XY(uint8_t Page, uint8_t Column);
void SED_pixel(uint8_t X, uint8_t Y, uint8_t pixelbyte);
unsigned int itoa( int z, char* Buffer );

//some Variables
volatile uint8_t data;							
volatile int pagecounter=0, columncounter=0;

//a little delay
void delay (void)
{
 volatile unsigned char a,b;
 for (a=0;a<=150;a++)
 	{
	 for (b=0;b<=150;b++)
	 	{
		 asm volatile("NOP");
	 	}
	 }
}


//Initialize SPI
void Init_SPI (void)				
{ 
 if(SOFT_SPI<=0)
 	{
	 LCDDIR |=(1<<MOSI)|(1<<SCK)|(1<<SS);	//Set SPI-output directions
 	 LCDPORT|=(1<<MOSI)|(1<<SCK);			//Set SPI-output
 	 SPCR|= (1<<MSTR);						//Set SPI-Master mode
 	 switch (SPICLK)
 		{
  		 case (2) : SPCR&=~(1<<SPR0); SPCR&=~(1<<SPR1); SPCR|= (1<<SPI2X); break;
  		 case (8) : SPCR|= (1<<SPR0); SPCR&=~(1<<SPR1); SPCR|= (1<<SPI2X); break;
  		 case (16): SPCR|= (1<<SPR0); SPCR&=~(1<<SPR1); SPCR&=~(1<<SPI2X); break;
  		 case (32): SPCR&=~(1<<SPR0); SPCR|= (1<<SPR1); SPCR|= (1<<SPI2X); break;
  		 case (64): SPCR&=~(1<<SPR0); SPCR|= (1<<SPR1); SPCR&=~(1<<SPI2X); break;
  		 case (128):SPCR|= (1<<SPR0); SPCR|= (1<<SPR1); SPCR&=~(1<<SPI2X); break;
  		 default:   SPCR&=~(1<<SPR0); SPCR&=~(1<<SPR1); SPCR&=~(1<<SPI2X); break;  //default is prescaler 4
 		}				
 	 SPCR&=~(1<<CPOL);						//Phase positiv
 	 SPCR&=~(1<<DORD);						//Data-Order MSB
 	 SPCR&=~(1<<SPIE);						//Interrupt not used
 	 SPCR|= (1<<CPHA);						//Clock-Phase High 
 	 SPCR|= (1<<SPE);						//Enable SPI
 	}
}

//send one byte over SPI
void SPI_out (uint8_t Data)					//Hardware SPI
{ 
 if(SOFT_SPI<=0)
	{
	 SPDR=Data;
 	 while(!(SPSR&(1<<SPIF)))
	 	{
		 asm volatile ("nop");
	 	}
	}
	else
 		{									//Software SPI-out
	 	unsigned char out;
	 	SOFT_SPI_PORT&=~(1<<SOFT_MOSI);
	 	SOFT_SPI_PORT&=~(1<<SOFT_SCK);
		 for(out=8; out>=1; out--)
	 		{
    	 	SOFT_SPI_PORT&=~(1<<SOFT_MOSI);  
	    	 if(Data & 0b10000000)
			 	{
      			 SOFT_SPI_PORT|=(1<<SOFT_MOSI); 
				} 
    	 	Data = Data<<1;
    	 	SOFT_SPI_PORT|= (1<<SOFT_SCK);
    	 	SOFT_SPI_PORT&=~(1<<SOFT_SCK); 
  			}
		}
}

//Initialize the SED1530 (E0855-2)
void Init_SED1530 (void)
{
 Init_SPI();								//Initialise the SPI
 RES_On;									// RESET the SED
 if (CS_ON>=1)
 	{
	 LCDPORT&=~(1<<CS1);					// /CS1 goes LOW
	 LCDPORT|= (1<<CS2);					// CS2 goes HI
	}
 RES_Off;									// SED is resetted
 delay();
 A0_Lo;										// Enter SED1530 command-mode
 SPI_out(0b01000000);						// Start display Line 0
 SPI_out(0b10100001);						// ADC inverse direction
 SPI_out(0b10100010);						// LCD-Bias= 1/5
 SPI_out(0b11000000);						// Output Status 0
 SPI_out(0b00101111);						// GLCD Voltage-Regulators ON
 delay();
 SPI_out(0b10001111);						// Contrast=Medium
 SPI_out(0b10100100);						// Display-Test OFF
 SPI_out(0b10101111);						// Display ON
 SPI_out(0b10100110);						// Display NOT inverted
 SPI_out(0b10110011);						// Set Page 3 (top-left)						
 SPI_out(COL_Hi);							// Set Column by Hi 4-Bits															
 SPI_out(COL_Lo);							// Set Column by Low 4-Bits
 A0_Hi;										// Enter SED1530 data-mode										 
}

// Clear Screen
void SED_clear (void)
{
 uint8_t pagecount;
 for (pagecount=0; pagecount<=8; pagecount++)		//Clear Display by erasing all Pixels
 	{
	 SED_Chg_page(pagecount);						//Set appropriate Page
	 SED_Home_Col();								//Set left column
	 for(columncounter=0;columncounter<SC_width;columncounter++)
	 	{
		 SPI_out(0x00);
		}
	}
 SED_Chg_page(0x03);								//Change to Page 3
 SED_Home_Col();									//Set left column
}

// Clear Page
void SED_clear_page (uint8_t Page)
{
 SED_Chg_page(Page);								//Set appropriate Page
 SED_Home_Col();									//Set left column
 for (columncounter=0;columncounter<SC_width;columncounter++)
 	{
	 SPI_out(0x00);
	}
 SED_Home_Col();									//Set left column
}

//Change Page
void SED_Chg_page (uint8_t Page)
{
 A0_Lo;
 switch (Page)
	{
	 case (0) : SPI_out(0b10110000); break;		//Set Page 0
	 case (1) : SPI_out(0b10110001); break;		//Set Page 1
	 case (2) : SPI_out(0b10110010); break;		//Set Page 2
	 case (3) : SPI_out(0b10110011); break;		//Set Page 3
	 case (4) : SPI_out(0b10110100); break;		//Set Page 0
	 case (5) : SPI_out(0b10110101); break;		//Set Page 1
	 case (6) : SPI_out(0b10110110); break;		//Set Page 2
	 case (7) : SPI_out(0b10110111); break;		//Set Page 3
	 case (8) : SPI_out(0b10111000); break;		//Set Page 8 (Special characters for The E0855-2)
	}
 pagecounter=Page;
 A0_Hi;
}


//Set the Left visible Corner of the Display
void SED_Home_Col (void)
{
 A0_Lo;											// Enter SED1530 command-mode
 SPI_out(COL_Hi);								// Set Column by Hi and Low 4-Bits
 SPI_out(COL_Lo);
 columncounter=0;
 A0_Hi;											// Enter SED1530 data-mode
}

//Display one Character
void SED_putc (char character)
{
 uint8_t column;	
 if (character>=32)
 	{
	 SPI_out(0x00);
	 columncounter++;
 	 for(column=0;column<=4;column++)
 		{
		 if (columncounter>=SC_width)
			{
			 SED_newline();	
			}
		 if (pagecounter<0)
			{	
			 pagecounter=3;
			 SED_Chg_page(pagecounter);
			 SED_Home_Col ();	
			}	 	
 //	 	 data=pgm_read_byte(&font[character-32][column]);		//Get charcter out of PROGMEM
 		 data=eeprom_read_byte(&font[character-32][column]);	//Get charcter out of EEPROM
 	 	 SPI_out(data);
		 columncounter++;
		}	
	}
}

//reverse the Bit-order for inverse Display of the E0855-2
//This is the only function, which is not my coding.
//Respect and honor to Mr. Huslik for this great function. 
//Not needed for displaying characters but still needed for graphics
uint8_t ReverseBitorder (uint8_t Bits)
{
  uint8_t ReversedBit;
  ReversedBit = (Bits >> 4) | (Bits << 4);
  ReversedBit = ((ReversedBit & 0xcc) >> 2 ) | ((ReversedBit & 0x33) << 2);
  ReversedBit = ((ReversedBit & 0xaa) >> 1 ) | ((ReversedBit & 0x55) << 1);
  return ReversedBit;
}

//display character-String
void SED_puts (char *string)
{
 while (*string)
 {
  SED_putc(*string);
  string++;
 }
}

//display character-String out of Progmem
void SED_puts_p (char *s_progmem)
{
 char p;
 while ((p=pgm_read_byte(s_progmem)))
 {
  SED_putc(p);
  s_progmem++;
 }
}

//display numeric-Values
void SED_puts_num (uint16_t numeric)
{
 int buffer[20];
 itoa(numeric,buffer);
 SED_puts(buffer);
}

//Show a picture out of the picture-file
void SED_showPIC (uint8_t picture)
{ 
 uint16_t column;
 SED_Chg_page(3);
 SED_Home_Col();
  	 for(column=0;column<=383;column++)
 		{
		 switch (picture)
 			{
	 		case (0): data=pgm_read_byte(&picture0[column]); break;	
	 		case (1): data=pgm_read_byte(&picture1[column]); break;
	 		case (2): data=pgm_read_byte(&picture2[column]); break;
	 		case (3): data=pgm_read_byte(&picture3[column]); break;
	 		case (4): data=pgm_read_byte(&picture4[column]); break;
	 		case (5): data=pgm_read_byte(&picture5[column]); break;
	 		case (6): data=pgm_read_byte(&picture6[column]); break;
	 		case (7): data=pgm_read_byte(&picture7[column]); break;
	 		case (8): data=pgm_read_byte(&picture8[column]); break;
	 		case (9): data=pgm_read_byte(&picture9[column]); break;
			}		 
		 data= ReverseBitorder(data);
 	 	 SPI_out(data);
		 switch (column)
		 	{
			 case (SC_width-1) : SED_Chg_page(2); SED_Home_Col(); break;
			 case ((SC_width-1)+SC_width): SED_Chg_page(1); SED_Home_Col(); break;
			 case ((SC_width-1)+(SC_width*2)): SED_Chg_page(0); SED_Home_Col(); break;
			}
		}
 SED_Chg_page(3);
 SED_Home_Col();	  
}

//Access to 7-segment-display (segno: most right=0, most left=3; number: 0 to 9)
void SED_7segment(uint8_t segno, uint8_t number)
{
 unsigned char temp,seg;
 switch (segno)								//Chose the desired 7-segment
 	{
	 case (0): segno=94; break;
	 case (1): segno=87; break;
	 case (2): segno=77; break;
	 case (3): segno=70; break;
	}
// seg=pgm_read_byte(&segment[number]);		//Read corresponding number-definition byte aut of PROGMEM
 seg=eeprom_read_byte(&segment[number]);  	//Read corresponding number-definition byte aut of EEPROM
 for (temp=0; temp<=6; temp++)
 	{
	 if(seg & 0x80)							//Compare last bit
	 	{
	 	 SED_symbol(segno-temp,1);			//On match display
		}
	    else
			{
	 	 	 SED_symbol(segno-temp,0);		//No match, no display
			}
     seg<<=1;								//Shift left
	} 
}

//Access to all the Symboles of the E2855-2
void SED_symbol(uint8_t symbol, uint8_t OnOff)
{
 unsigned char temp;
 temp=pagecounter;
 SED_set_XY(8,symbol-16);
 if (OnOff>=1)
 	{
	 SPI_out(0x01);
    }
	else
		{
		 SPI_out(0x00);
		}
 SED_Chg_page(temp);
 SED_Home_Col();
 }

//Newline and carriage return
void SED_newline()
{
 pagecounter--;
 	if (pagecounter<0)
	{
	 pagecounter=3;
	 SED_Chg_page(pagecounter);
	 SED_Home_Col ();	
	}
	else
	{
 	 SED_Chg_page(pagecounter);
 	 SED_Home_Col ();
	} 
}

//Select Page and Column
void SED_set_XY(uint8_t Page, uint8_t Column)
{
 SED_Home_Col ();
 SED_Chg_page (Page);
 Column=Column+16;
 volatile unsigned char col_HI=Column,col_LO=Column;
 columncounter=Column-16;
 A0_Lo;										
 col_HI=col_HI>>4;
 col_LO=col_LO<<4;
 col_LO=col_LO>>4;
 col_HI=col_HI+16;
 SPI_out(col_HI);
 SPI_out(col_LO);
 A0_Hi;
}

//Send Display to sleep
void SED_sleep (void)
{
 A0_Lo;
 SPI_out(0b10101101);				//Static indicator ON
 SPI_out(0b00101000);				//Turn OFF coltage follower, regulator and booster
 A0_Hi;

}

// wake-up Display from sleep
void SED_wakeup (void)
{
 A0_Lo;
 SPI_out(0b10101101);				//Static indicator OFF
 SPI_out(0b00101111);				//Turn ON coltage follower, regulator and booster
 delay();
 A0_Hi;
}

//Set or change the contrast of the display in 32 (0-31) steps
void SED_contrast (uint8_t contrast)
{ 
 A0_Lo;
 if (contrast<=31)
 	{
	 contrast=contrast+128;
 	 SPI_out(contrast);				// Range is: 0 to 31
	}
	else
		{
		 SPI_out(0b10001111);		// Contrast=Medium=15
		}
 A0_Hi;
}

//Set or unset a Pixel (Parameters: column[0-95], Page[0-3], Bytepattern)
void SED_pixel(uint8_t X, uint8_t Y, uint8_t pixelbyte)
{
 SED_set_XY(Y,X);					//SED_set_XY takes (Page/Column)
 SPI_out(pixelbyte);				//Send the vertical byte-pattern to the according Page
} 

//itoa-function (taken from www.mikrocontroller.net/articles/FAQ#itoa.28.29)
unsigned int itoa( int z, char* Buffer )
{
  int i = 0;
  int j;
  char tmp;
  if( z == 0 ) {
      Buffer[0] = '0';
      Buffer[1] = '\0';
  }
  else 
	{
     if( z < 0 ) 
		{
		 Buffer[0] = '-';
		 z = -z;
		 Buffer++;
		}
     while( z > 0 )
		{
		 Buffer[i++] = '0' + z % 10;
		 z /= 10;
		}
    for( j = 0; j < i / 2; ++j ) 
		{
		 tmp = Buffer[j];
		 Buffer[j] = Buffer[i-j-1];
		 Buffer[i-j-1] = tmp;
		} 
		Buffer[i] = '\0';		
	}
return Buffer;
}
